import { promises as fs, existsSync } from "fs"
import { Buffer } from "buffer"

import type { IAdapter, IFileSystem } from "./index"

function splitPath(path: string) {
  const segments = path.split("/")
  const filename = segments.pop()!
  return [segments, filename] as const
}

export class BrowserAdapter implements IFileSystem {
  constructor(private handle: FileSystemDirectoryHandle) {}

  static async create() {
    const handle = await showDirectoryPicker()
    return new BrowserAdapter(handle)
  }

  private async getDirectoryHandle([first, ...pathSegments]: string[]) {
    const handle = await pathSegments.reduce(
      async (accum, next) => (await accum).getDirectoryHandle(next),
      this.handle.getDirectoryHandle(first)
    )
    return handle
  }

  private async getFileHandle(path: string) {
    const [segments, filename] = splitPath(path)
    const dirHandle = await this.getDirectoryHandle(segments)
    const fileHandle = await dirHandle.getFileHandle(filename)
    return fileHandle
  }

  async readFile(path: string) {
    const handle = await this.getFileHandle(path)
    const file = await handle.getFile()
    return file.text()
  }

  async existsSync(path: string): Promise<boolean> {
    const [segments, filename] = splitPath(path)
    let handle = this.handle
    for (const segment of segments) {
      handle.values()
      const next = await handle.getDirectoryHandle()
    }

    throw new Error("Method not implemented.")
  }

  async readBuffer(path: string): Promise<Buffer> {
    const handle = await this.getFileHandle(path)
    const file = await handle.getFile()
    return Buffer.from(await file.arrayBuffer())
  }

  async writeFile(path: string, data: string): Promise<void> {
    const handle = await this.getFileHandle(path)
    const writable = await handle.createWritable()
    await writable.write(data)
    await writable.close()
  }

  readdir(path: string): Promise<string[]> {
    throw new Error("Method not implemented.")
  }

  stat(path: string): Promise<{ isDirectory(): boolean }> {
    throw new Error("Method not implemented.")
  }
}

/**
 * Default Node.js adapter. This can be used while using `opvault.js`
 * in a Node.js environment.
 */
const nodeAdapter: IAdapter = {
  fs: {
    readFile: path => fs.readFile(path, "utf-8"),
    readBuffer: path => fs.readFile(path),
    writeFile: fs.writeFile,
    readdir: fs.readdir,
    stat: fs.stat,
    existsSync,
  },
}

export default nodeAdapter
