import type { Crypto } from "./crypto"
import { decryptOPData } from "./crypto"
import { invariant } from "../errors"

type integer = number

export interface AttachmentMetadata {
  itemUUID: string
  contentSize: integer
  external: boolean
  updatedAt: integer
  txTimestamp: integer
  /** Encoded as base64 encoded OPData */
  overview: {
    filename: string
  }
  createdAt: integer
  uuid: string
}

export class Attachment {
  #k: string
  #crypto: Crypto
  #buffer: Buffer

  #icon?: Buffer // png buffer
  #file?: Buffer
  #metadata?: AttachmentMetadata

  private metadataSize: number
  private iconSize: number

  constructor(crypto: Crypto, k: string, buffer: Buffer) {
    this.#buffer = buffer
    this.#validate()
    this.#crypto = crypto
    this.#k = k
    this.metadataSize = buffer.readIntLE(8, 2)
    this.iconSize = buffer.readIntLE(12, 3)

    crypto.on("lock", () => {
      this.#lock()
    })
  }

  /**
   * Validate attachment file.
   */
  #validate() {
    const file = this.#buffer
    invariant(
      file.slice(0, 6).toString("utf-8") === "OPCLDA",
      "Attachment must start with OPCLDA"
    )
    invariant(
      file.readIntLE(7, 1) === 1,
      "The version for this attachment file format is not supported."
    )
  }

  get icon() {
    if (this.#icon == null) {
      this.#decrypt()
    }
    return this.#icon!
  }

  get file() {
    if (this.#file == null) {
      this.#decrypt()
    }
    return this.#file!
  }

  get metadata() {
    if (this.#metadata == null) {
      this.#decrypt()
    }
    return this.#metadata!
  }

  #decrypt() {
    const cipher = this.#crypto.deriveConcreteKey({ k: this.#k })
    const { metadataSize, iconSize } = this
    const buffer = this.#buffer
    this.#icon = decryptOPData(
      buffer.slice(16 + metadataSize, 16 + metadataSize + iconSize),
      cipher
    )
    this.#file = decryptOPData(buffer.slice(16 + metadataSize + iconSize), cipher)

    const metadata = JSON.parse(buffer.slice(16, 16 + metadataSize).toString("utf-8"))
    metadata.overview = JSON.parse(
      decryptOPData(Buffer.from(metadata.overview, "base64"), this.#crypto.ov).toString()
    )
    this.#metadata = metadata
  }

  #lock() {
    this.#metadata = undefined!
    this.#icon = undefined!
    this.#file = undefined!
  }

  dispose() {
    this.#buffer = null!
    this.#lock()
  }
}
