import * as crypto from "crypto"
import { createHmac, createDecipheriv, createHash } from "crypto"
import { EventEmitter } from "../ee"
import { HMACAssertionError } from "../errors"
import type { i18n } from "../i18n"
import type { ItemDetails, Overview, Profile } from "../types"
import { setIfAbsent } from "../util"
import type { EncryptedItem } from "./item"

/** Encryption and MAC */
export interface Cipher {
  /** Encryption key */
  key: Buffer
  /** HMAC key */
  hmac: Buffer
}

export class Crypto extends EventEmitter<{ lock: void }> implements IDisposable {
  #disposables: IDisposable[] = []
  #locked = true

  #master!: Cipher
  #overview!: Cipher

  constructor(readonly i18n: i18n) {
    super()
  }

  unlock(profile: Profile, masterPassword: string) {
    const derivedKey = crypto.pbkdf2Sync(
      masterPassword,
      Buffer.from(profile.salt, "base64"),
      profile.iterations,
      64,
      "sha512"
    )

    const cipher = splitPlainText(derivedKey)

    // Derive master key and overview keys
    this.#master = decryptKeys(profile.masterKey, cipher)
    this.#overview = decryptKeys(profile.overviewKey, cipher)
    this.#locked = false
  }

  get locked() {
    return this.#locked
  }

  lock() {
    this.#locked = true
    this.#master = null!
    this.#overview = null!
    this.#disposables.forEach(fn => fn.dispose())
    this.emit("lock")
  }

  dispose() {
    this.lock()
  }

  assertUnlocked() {
    if (this.#locked) {
      throw new Error(this.i18n.error.vaultIsLocked)
    }
  }

  #createCache = <K, V, K2 = K>(
    deriveArg: (value: K) => K2,
    implementation: (value: K2) => V
  ) => {
    const map = new Map<K2, V>()
    this.#disposables.push({
      dispose: () => map.clear(),
    })
    return (data: K) => setIfAbsent(map, deriveArg(data), implementation)
  }

  #createWeakCache = <K extends object, V>(implementation: (value: K) => V) => {
    let map = new WeakMap<K, V>()
    this.#disposables.push({
      dispose() {
        map = new WeakMap()
      },
    })
    return (data: K) => setIfAbsent(map, data, implementation)
  }

  decryptItemDetails = this.#createWeakCache((item: EncryptedItem) => {
    const cipher = this.deriveConcreteKey(item)
    const detail = decryptOPData(Buffer.from(item.d, "base64"), cipher)
    return JSON.parse(detail.toString("utf-8")) as ItemDetails
  })

  decryptItemOverview = this.#createCache(
    (item: EncryptedItem) => item.o,
    (o: string) => {
      const overview = decryptOPData(Buffer.from(o, "base64"), this.#overview)
      return JSON.parse(overview.toString("utf8")) as Overview
    }
  )

  deriveConcreteKey = this.#createCache(
    (data: { k: string }) => data.k,
    ($k: string) => {
      const k = Buffer.from($k, "base64")
      const data = k.slice(0, -32)
      assertHMac(data, this.#master.hmac, k.slice(-32))
      const derivedKey = decryptData(this.#master.key, data.slice(0, 16), data.slice(16))
      return splitPlainText(derivedKey)
    }
  )

  deriveGeneralKey = (input: Buffer) => {}

  get ov() {
    return this.#overview
  }
}

// async function pbkdf2(password: string, salt: string, iterations = 1000, length = 256) {
//   const encoder = new TextEncoder()
//   const key = await webcrypto.subtle.importKey(
//     "raw",
//     encoder.encode(password),
//     "PBKDF2",
//     false,
//     ["deriveBits"]
//   )
//   const bits = await webcrypto.subtle.deriveBits(
//     {
//       name: "PBKDF2",
//       hash: "SHA-512",
//       salt: encoder.encode(salt),
//       iterations,
//     },
//     key,
//     length
//   )
//   return bits
// }

export function assertHMac(data: Buffer, key: Buffer, expected: Buffer) {
  const actual = createHmac("sha256", key).update(data).digest()
  if (!actual.equals(expected)) {
    console.error({ actual, expected })
    throw new HMACAssertionError()
  }
}

export const splitPlainText = (derivedKey: Buffer): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

export function decryptKeys(encryptedKey: string, derived: Cipher) {
  const buffer = Buffer.from(encryptedKey, "base64")
  const base = decryptOPData(buffer, derived)
  const digest = createHash("sha512").update(base).digest()
  return splitPlainText(digest)
}

export function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  const cipher = createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false)
  return Buffer.concat([cipher.update(data), cipher.final()])
}

export function decryptOPData(cipherText: Buffer, cipher: Cipher) {
  const key = cipherText.slice(0, -32)
  assertHMac(key, cipher.hmac, cipherText.slice(-32))

  const plaintext = decryptData(cipher.key, key.slice(16, 32), key.slice(32))
  const size = readUint16(key.slice(8, 16))
  return plaintext.slice(-size)
}

function readUint16({ buffer, byteOffset, length }: Buffer) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}
